/*
 Wireless Gameboy/GBA Controller - Nintendo 64 RX
 Version: 1.2
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 13/03/2019
 Last Modified: 30/05/2019
 
 Use any Gameboy / GBA which supports a Gameboy or GBA cartridge such as the Gameboy, GBC, GBP, GBA, GBA SP, 
 DS, DS Lite and GB Micro as a Wireless game controller for any device which supports USB HID Joysticks / Gamepad.
 
 This version now supports the Multi-TX boards.
 
 Program the ATtiny441/841 to use a 12MHz crystal
 avrdude -p ATtiny441 -c usbasp -U lfuse:w:0xff:m -U hfuse:w:0xdf
 
*/

// ATtiny441/841 Pin map
//
//                     +-\/-+
//               VCC  1|o   |14  GND
//    16MHz XTAL PB0  2|    |13  PA0 Input
//    16MHz XTAL PB1  3|    |12  PA1 Output
//         Reset PB3  4|    |11  PA2 LED
//               PB2  5|    |10  PA3 nRF24L01 CE
//  nRF24L01 CSN PA7  6|    |9   PA4 nRF24L01 SCK
//   nRF24L01 DI PA6  7|    |8   PA5 nRF24L01 DO
//                     +----+

#define F_CPU 12000000

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/sleep.h>
#include <math.h>
#include <util/delay.h>
#include "setup.c"
#include "mirf.c"
#include "transmitters.c"

uint8_t modeFound = 0;
uint8_t kpData = 0;
uint8_t kpDataHigh = 0;
uint8_t firstPacket = 1;
uint8_t kpOutputDpad = 0;
uint8_t kpOutputMapYB = 0;
uint8_t kpOutputMapLtoZ = 0;
uint8_t volatile timedOut = 0;
uint8_t multiTxData[10];

uint8_t tempKpOutput[4] = {0x00, 0x00, 0x00, 0x00};
uint8_t volatile n64ControllerData[4] = {0x00, 0x00, 0x00, 0x00};

int main (void) {
	setup();
	
	check_for_new_settings(); // Check for new nRF settings
	check_for_tx_mode(); // Check for TX mode
	
	// Turn on interrupt
	PCMSK0 |= (1<<PCINT0);
	GIMSK |= (1<<PCIE0);
	
	// Turn on time out counter for ~5 seconds
	TCNT1 = 0;
	TIMSK1 = (1<<TOV1);
	TCCR1B = (1<<CS12) | (1<<CS10);
	
	while(1) {
		// If timed out, reset to defaults and check for the transmitter again
		if (timedOut == 1) {
			n64ControllerData[0] = 0x00;
			n64ControllerData[1] = 0x00;
			n64ControllerData[2] = 0x00;
			n64ControllerData[3] = 0x00;
			
			check_for_tx_mode();
			timedOut = 0;
			TCNT1 = 0;
			TCCR1B = (1<<CS12) | (1<<CS10);
		}
		
		// Packet received
		if (mirf_status() & (1<<RX_DR)) {
			PORTA |= (1<<LED);
			mirf_CE_lo; // Stop listening
			
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			
			if (modeFound == SNES_TX) { // Read 3 bytes if in SNES mode
				for (uint8_t x = 0; x < 3; x++) {
					multiTxData[x] = spi_transfer(0);
				}
			}
			else if (modeFound == GBA_MODE) { // Load 2 bytes in GBA mode
				kpDataHigh = spi_transfer(0);
				kpData = spi_transfer(0);
			}
			else {
				kpData = spi_transfer(0); // Load 1 byte in GB mode
			}
			
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			mirf_flush_rx_tx(); // Flush TX/RX
			mirf_CE_hi; // Start listening
			PORTA &= ~(1<<LED);
			
			// Reset timeout timer
			TCNT1 = 0;
			
			// Reset temp output
			tempKpOutput[0] = 0;
			tempKpOutput[1] = 0;
			tempKpOutput[2] = 0;
			tempKpOutput[3] = 0;
			
			// Action based on TX found
			if (modeFound == SNES_TX) {
				tx_snes();
			}
			else {
				tx_gb_gba();
			}
			
			// Assign new key presses
			n64ControllerData[0] = tempKpOutput[0];
			n64ControllerData[1] = tempKpOutput[1];
			n64ControllerData[2] = tempKpOutput[2];
			n64ControllerData[3] = tempKpOutput[3];
		}
		
		_delay_us(250);
	}
	
	return 0;
}

// Data input was toggled
ISR(PCINT0_vect) {
	if (!(PINA & (1<<DATA_IN))) {
		PORTA &= ~(1<<LED);
		
		uint8_t dataIn = 0;
		for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
			uint8_t bitTime = 0;
			
			while (!(PINA & (1<<DATA_IN))); // Wait until high
			//PORTA |= (1<<LED);
			asm volatile("nop");
			
			while (PINA & (1<<DATA_IN)) { // While high, count
				bitTime++;
			}
			//PORTA &= ~(1<<LED);
			asm volatile("nop");
			
			if (bitTime >= 6) { // High
				dataIn |= x;
			}
		}
		
		while (!(PINA & (1<<DATA_IN))); // Wait until high
		_delay_us(1);
		
		// Write the controller data
		for (uint8_t byte = 0; byte < 4; byte++) {
			for (uint8_t x = 0x80; x != 0; x >>= 1) { // Loop 8 times (fast)
				if (n64ControllerData[byte] & x) {
					PORTA |= (1<<DATA_OUT);
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					asm volatile("nop");
					PORTA &= ~(1<<DATA_OUT);
					_delay_us(3);
					PORTA |= (1<<DATA_OUT);
				}
				else {
					PORTA |= (1<<DATA_OUT);
					_delay_us(2);
					PORTA &= ~(1<<DATA_OUT);
					_delay_us(1);
					PORTA |= (1<<DATA_OUT);
				}
			}
		}
		PORTA |= (1<<DATA_OUT);
		_delay_us(2);
		PORTA &= ~(1<<DATA_OUT);
		
		_delay_us(5);
	}
}

// Overflows after ~5 seconds of no packets received
ISR(TIMER1_OVF_vect) {
	timedOut = 1;
	TCNT1 = 0;
	TCCR1B = 0;
}
